
import sys
from bl.dict import Dict

class RecordSet(list):
    """A set of database records"""

    def select(self, **args):
        """return all records in the set that have the attributes specified in args, or empty RecordSet"""
        result = RecordSet()
        for record in self:
            match = True
            for k in args:
                if record[k] != args[k]:
                    match = False
                    break
            if match == True:
                result.append(record)
        return result
        
    def select_one(self, **args):
        """return the first record in the set that has the attributes specified in args, or None."""
        for record in self:
            match = True
            for k in args:
                if record[k] != args[k]:
                    match = False
                    break
            if match == True:
                return record

    def dict(self, key):
        """returns the RecordSet as a dict of records, keyed to strings or tuples. 
            keys must be unique, or later duplicates will overwrite earlier ones."""
        d = Dict()
        for record in self:
            if type(key) == str:            # string key
                d[record[key]] = record
            else:                           # tuple key
                d[record.key_tuple(key)] = record
        return d
        
    def json(self, fields=[]):
        """Return the contents of this record set as json"""
        l = [record.json(fields=fields) for record in self]
        return '[' + ', '.join(l) + ']'

        





import datetime
from bl.dict import Dict
from bl.string import String

class Record(Dict):
    """A record from the database."""

    def __init__(self, db, **args):
        self.__dict__['db'] = db
        Dict.__init__(self, **args)

    def key_tuple(self, key=[]):
        """can be used as a dict key."""
        if key==[]: 
            key = self.keys()
        return tuple([self[k] for k in key])
        
    def json(self, fields=None, indent=None, tag=None):
        """Return the contents of this record as json"""
        import json as _json
        d = Dict()
        if tag is None:
            tag = String(self.__class__.__name__).identifier().lower()
        j = Dict(**{tag: d})
        for k in fields or list(self.keys()): 
            d[k] = self.get(k)
            if type(d[k]) == datetime.datetime:         # datetime not compatible with json.dumps
                d[k] = d[k].isoformat(' ')              # -- iso formatted string
        return _json.dumps(j, indent=indent)

    def using_keys(self, keys=[]):
        """return a copy of this record that only has the given keys. 
        Useful for generating a keyword-argument dictionary for select()
        """
        d = self.__class__(self.db)
        for k in keys:
            d[k] = self[k]
        return d

    def using_prefix(self, prefix, **args):
        """return an instance containing the args that have the given prefix, with prefix removed"""
        r = self.__class__.__init__(self, self.db)
        for k in [k for k in args.keys()
                if k[:len(prefix)]==prefix]:
            r[k[len(prefix):]] = args[k]
        return r





DEBUG = True

import datetime
import datetime, re, sys
from bl.dict import Dict
from bl.string import String

from .record import Record
from .recordset import RecordSet

class Model(Record):
    """abstract base class for database models to inherit from, holds shared functionality.

    # -- Use an in-memory sqlite3 database for testing: --
    >>> from bsql.database import Database
    >>> db = Database()
    >>> db.execute("create table model_test (id integer primary key autoincrement, name varchar)")

    # -- Example usage: --
    >>> class Temp(Model):                      # set up a temporary model
    ...     relation = 'model_test'
    ...     pk = ['id']
    ...
    >>> t = Temp(db)                            # new instance
    >>> t.name = 'something temporary'          # set an attribute
    >>> t.insert()                              # insert a new record
    >>> t                                       # values are reloaded from the insert
    {'id': 1, 'name': 'something temporary'}
    >>> t.name = "another thing"                # edit an attribute
    >>> t.commit()                              # update the record in the database
    >>> t
    {'id': 1, 'name': 'another thing'}
    >>> t1 = t.select_one(id=t.id)              # get the full record
    >>> t1
    {'id': 1, 'name': 'another thing'}
    >>> t2 = Temp(db)
    >>> t2.name = 'thing 2'
    >>> t2.insert()
    >>> ts = t.select(); len(ts)                # there are two records
    2
    >>> ts = t.select(name='thing 2')
    >>> len(ts)                                 # only one such record
    1
    >>> t1.delete()                             # delete a record
    >>> ts = t.select(); len(ts)                # there is now one record
    1

    -- The rest is teardown: --
    >>> db.execute("drop table model_test")
    """

    relation = None     # the database relation with which this model deals primarily
    pk = ['id']         # fields that make up the primary key of a record
                        #   is there a way to do this by introspection at instance init?
    where = None        # only include records that match this where clause.
                        #   often used to have multiple classes in one table (i.e., single-table inheritance)

    def __init__(self, db, **args):
        Record.__init__(self, db, **args)
        if self.relation is None: 
            self.__dict__['relation'] = String(self.__class__.__name__).identifier().lower() + 's'

    def to_one(self, other_class, to_fields=['*'], foreign_key=None, other_key=None, 
                     update=False, cache_field=None, orderby=None, **kwargs):
        """returns a record based on the fk fields in this relation."""
        # make sure we have a valid record
        for k in self.pk:
            if self.get(k) is None: return None
        
        if foreign_key is None: 
            foreign_key = ["%s_%s" % (other_class.__name__.lower(), fk) for fk in other_class.pk]
        if other_key is None: other_key = other_class.pk
        if cache_field is None: cache_field = other_class.__name__

        for k in foreign_key:
            if self.get(k) is None: return None

        if update==True or self.__dict__.get(cache_field)==None:
            # get the record from the db
            wherelist = []
            if self.where is not None: 
                wherelist += [self.where]
            fields = list(zip(foreign_key, other_key))
            for field in fields:
                wherelist.append("%s=%s" % (field[1], self.quote(self[field[0]])))
            wherecl = " and ".join(wherelist)
            attr = ','.join(to_fields)
            if orderby is not None:
                selexpr = """other_class(self.db).select_one(attr="%s", where="%s", orderby="%s", **kwargs)""" % (attr, wherecl, orderby)
            else:
                selexpr = """other_class(self.db).select_one(attr="%s", where="%s", **kwargs)""" % (attr, wherecl)
            self.__dict__[cache_field] = eval(selexpr)
        return self.__dict__[cache_field]

    def to_many(self, other_class, 
                self_key=None, foreign_key=None, update=False, cache_field=None, 
                orderby=None, where=None, **kwargs):
        """return a set of records from a foreign class that key to this instance."""
        # make sure we have a valid record
        for k in self.pk:
            if self.get(k) is None: return RecordSet()
        
        if self_key is None: self_key = self.pk
        if foreign_key is None: 
            foreign_key = ["%s_%s" % (self.__class__.__name__.lower(), fk) for fk in self.pk]
        if cache_field is None: cache_field = other_class.__name__

        for k in self_key:
            if self.get(k) is None: return RecordSet()

        if update==True or self.__dict__.get(cache_field)==None:
            # get the records from the database, however many there are
            wherelist = []
            if self.where is not None: wherelist += [self.where]
            fields = list(zip(self_key, foreign_key))
            for field in fields:
                wherelist.append(" %s=%s " % (field[1], self.quote(self[field[0]])))
            wherecl = " and ".join(wherelist)
            if where not in ['', None]: 
                wherecl += " and %s " % self.where_from_args(where)            
            if orderby is not None:
                selexpr = """other_class(self.db).select(where="%s", orderby="%s", **kwargs)""" % (wherecl, orderby)
            else:
                selexpr = """other_class(self.db).select(where="%s", **kwargs)""" % (wherecl)
            self.__dict__[cache_field] = eval(selexpr) or RecordSet()
        rs = RecordSet()
        for r in self.__dict__[cache_field] or []:
            rs.append(r)
        return rs

    def to_many_through(self, other_class, through_relation, through_fields, 
                        self_key=None, foreign_key=None, to_fields=['*'], 
                        also_select=[], cache_field=None, update=False, 
                        orderby=None, where=None, limit=None, offset=0, **kwargs):
        """return a set of records from a foreign class that key to this instance, 
        through a relationship table."""
        # make sure we have a valid record
        for k in self.pk:
            if self.get(k) is None: return RecordSet()
        
        if self_key is None: self_key = self.pk
        if foreign_key is None: 
            foreign_key = ["%s_%s" % (self.__class__.__name__.lower(), fk) for fk in self.pk]
        if cache_field is None: cache_field = other_class.__name__
        
        for k in self_key:
            if self.get(k) is None: return RecordSet()

        if update==True or self.__dict__.get(cache_field)==None:
            # add the relation names to the field names in each field list -- avoid ambiguity in the query
            self_key = ["%s.%s" % (self.relation, field) for field in self_key]
            foreign_key = ["%s.%s" % (other_class.relation, field) for field in foreign_key]
            through_fields = ["%s.%s" % (through_relation, field) for field in through_fields]

            # create lists for the inner joins
            onself = [" %s=%s " % field for field in zip(self_key, through_fields)]                      # uses the first set of fields in through_fields
            onforeign = [" %s=%s " % field for field in zip(foreign_key, through_fields[len(onself):])]  # uses the rest of the fields in through_fields

            # build the query
            sql = "select " + ", ".join(["%s.%s" % (other_class.relation, field) for field in to_fields])
            if type(also_select)==list and len(also_select) > 0:
                sql += ", " + ', '.join(also_select)
            elif type(also_select)==str and len(also_select)> 0:
                sql += ', ' + also_select
            sql+= " from %s" % other_class.relation
            sql+= "\n inner join %s on %s" % (through_relation, " and ".join(onforeign))
            sql+= "\n inner join %s on %s" % (self.relation, " and ".join(onself))
            sql+= "\n where %s" % " and ".join(["%s=%s" % (field, self.quote(self[field.split('.')[1]])) for field in self_key])
            if where not in ['', None] or len(kwargs)>0: 
                sql += " and %s " % self.where_from_args(where)
            if orderby in ['', None]: orderby = ",".join(other_class.pk)
            sql += "\n order by %s" % orderby            
            if limit not in [0, None]: sql += " limit %d" % int(limit)
            if offset != 0 and type(offset) == int:
                sql += " offset %d " % offset
            
            # select and cache the data
            self.__dict__[cache_field] = self.db.elect(sql, Record=other_class)
        rs = RecordSet()
        for r in self.__dict__[cache_field] or RecordSet():
            rs.append(r)
        return rs

    def prepare_query(self, attr='*', from_expr=None, where="", vals=None, orderby="", limit=None, offset=0, **kwargs):
        if from_expr is None: from_expr = self.relation
        if limit not in [0, None] and self.db.servername() == 'sqlserver':
            sql = "select top %d %s from %s" % (limit, attr, from_expr)
        else:
            sql = "select %s from %s" % (attr, from_expr)
        if where not in ["", None] or len(kwargs)>0: 
            sql += " where %s " % self.where_from_args(where, **kwargs)

        if orderby not in ["", None]:
            sql += """ order by %s """ % orderby
        elif len(self.pk) > 0:       # default to using pk for orderby
            sql += """ order by %s """ % '","'.join(self.pk)
        if limit not in [0, None] and self.db.servername() != 'sqlserver': 
            sql += " limit %d" % int(limit)

        if offset != 0:
            if type(offset) == int:
                sql += " offset %d " % offset
            elif type(offset) in [str, str] and re.match("[0-9]+", offset) is not None:
                sql += " offset %s " % offset
        return sql

    def selectgen(self, attr='*', from_expr=None, where="", vals=None, orderby="", limit=None, offset=0, cursor=None, **kwargs):
        """select records from relation"""
        sql = self.prepare_query(attr=attr, from_expr=from_expr, where=where, vals=vals, orderby=orderby, limit=limit, offset=offset, **kwargs)
        results = self.db.selectgen(sql, vals=vals, Record=self.__class__, cursor=cursor)
        for result in results:
            result.after_select()
            yield result

    def select(self, attr='*', from_expr=None, where="", vals=None, orderby="", limit=None, offset=0, cursor=None, **kwargs):
        """select records from relation"""
        sql = self.prepare_query(attr=attr, from_expr=from_expr, where=where, vals=vals, orderby=orderby, limit=limit, offset=offset, **kwargs)
        records = self.db.selectgen(sql, vals=vals, Record=self.__class__, cursor=cursor)
        results = RecordSet()
        for result in records:
            result.after_select()
            results.append(result)
        return results

    def select_one(self, attr='*', from_expr=None, where="", vals=None, orderby="", offset=0, cursor=None, **kwargs):
        """select one record from relation"""
        results = self.selectgen(attr=attr, from_expr=from_expr, where=where, vals=vals, orderby=orderby, limit=1, offset=offset, **kwargs)
        try:
            result = results.__next__()
        except StopIteration:
            result = None
        return result

    def select_as_dict(self, attr='*', from_expr=None, where="", vals=None, orderby="", cursor=None, **kwargs):
        """returns the SELECT results as a dict, with the keys being tuples of the model's pk."""
        records = self.select(attr=attr, from_expr=from_expr, where=where, vals=vals, orderby=orderby, cursor=cursor, **kwargs)
        d = Dict()
        for record in records:
            d[record.pk_as_tuple()] = record
        return d
            
    def where_from_args(self, where=None, **kwargs):
        wherelist=[]
        if where!='' and where is not None: wherelist.append(where)
        for k in kwargs:
            wherelist.append("%s=%s" % (k, self.quote(kwargs[k])))
        return " and ".join(wherelist)  # returns empty string if no kwargs given

    def pk_as_tuple(self):
        """return a tuple with the values of self.pk -- can be used as a dict key."""
        key=()  
        for k in self.pk:
            key = key + (self[k],)
        return key

    def execute(self, sql, vals=None, cursor=None):
        return self.db.execute(sql, vals=vals, cursor=cursor)

    def exists(self, cursor=None):
        return self.select_one(cursor=cursor, 
            **{k:self[k] for k in self.pk}) is not None

    def insert(self, reload=True, cursor=None, **kwargs):
        """insert the current instance into its relation."""
        for k in list(kwargs.keys()): self[k] = kwargs[k]
        self.before_insert_or_update()
        self.before_insert()
        keys = self.keys()
        vals = self.values()
        if 'postgres' in self.db.servername().lower():
            vals_markers = ["%s" for v in vals]
        else:
            vals_markers = ["?" for v in vals]
        q = "insert into %s (%s) values (%s)" % (
                self.relation, ','.join(keys), ','.join(vals_markers))
        if 'postgres' in self.db.servername().lower(): 
            q += " returning *"
            d = self.db.select_one(q, vals=vals, cursor=cursor)
        else:
            d = self.db.execute(q, vals=vals, cursor=cursor)
        if reload==True:
            if 'sqlite' in self.db.servername().lower():
                d = self.db.select_one("select * from %s where ROWID=last_insert_rowid()" % self.relation)
            elif 'postgres' not in self.db.servername().lower() \
            and None not in [self.get(k) for k in self.pk]:    # local pk is filled
                whereargs = {}
                for k in self.pk:
                    whereargs[k] = self.get(k)
                wherecl = self.where_from_args(**whereargs)
                d = self.db.select_one("select * from %s where %s" % (self.relation, wherecl))
            for k in list(d.keys()): self[k] = d[k]
        if cursor is None:
            self.db.commit()
        self.after_insert()
        self.after_insert_or_update()

    def insert_safe(self, cursor=None, **args):
        try:
            self.insert(**args)
        except:
            return sys.exc_info()[1].args[0]        

    def insert_if_none(self, reload=True, cursor=None):
        """insert the current instance if it's not there."""
        if self.select_one(cursor=cursor, **self) is None:
            return self.insert(reload=reload)

    def reload(self, **kwargs):
        """reload the current instance from the database."""
        al = []
        if kwargs != {}:
            d = kwargs
            for k in list(d.keys()):
                al.append("%s=%s" % (k, self.quote(d[k])))
        else:
            for k in self.pk:
                al.append("%s=%s" % (k, self.quote(self[k])))
        i = self.select_one(where=" and ".join(al))
        if i is not None:
            for k in list(i.keys()):
                self[k] = i[k]
        else:
            del(self)

    def commit(self, reload=True, fields=[], cursor=None, **kwargs):
        """update this instance's record in its relation."""

        self.before_insert_or_update()
        self.before_update()

        vals = []

        # update the local copy of any attribute with the given arg.
        for k in list(kwargs.keys()):
            self[k] = kwargs[k]

        # make attribute list
        al = []
        if fields == []:
            keys = list(self.keys())
        else:
            keys = fields 
        
        for k in keys:
            # By default, only update fields that are not pk.
            # But if the fields to update have been specified, try updating all of them.
            if k not in self.pk or fields != []:
                al.append("%s=%%s" % (k))     # append to attribute list
                vals.append(self[k])

        # join pk to indicate which record to update.
        pl = []
        for k in self.pk:
            pl.append("%s=%%s" % (k))
            vals.append(self[k])
        where = ' and '.join(pl)

        # perform the update
        sql = "update %s set %s where %s" % (self.relation, ', '.join(al), where)
        self.db.execute(sql, vals=vals, cursor=cursor)
        
        self.after_update()
        self.after_insert_or_update()
        if reload==True: self.reload()

    def commit_safe(self, cursor=None, **args):
        try:
            self.commit(cursor=cursor, **args)
        except:
            return sys.exc_info()[1].args[0]        

    def insert_or_update(self, reload=True, cursor=None, **kwargs):
        pksel = {k:self.get(k) for k in self.pk}
        if None not in pksel.values() and self.select_one(**pksel) is not None:
            self.commit(reload=reload, cursor=cursor, **kwargs)
        else:
            self.insert(reload=reload, cursor=cursor, **kwargs)

    def delete(self, where=None, vals=None, cursor=None):
        """delete the current instance from its relation -- must be up-to-date to do w/o where clause."""
        self.before_delete()
        if where==None:
            # use pk to find the appropriate record
            l=[]
            for k in self.pk:
                l.append("%s=%s" % (k, self.quote(self[k])))
            where=" and ".join(l)
        self.db.execute("delete from %s where %s" % (self.relation, where), vals=vals, cursor=cursor)
        for k in list(self.keys()):
            self.pop(k)
        self.after_delete()

    def delete_safe(self, where=None, vals=None, cursor=None, **args):
        try:
            self.delete(where=where, vals=vals, cursor=cursor, **args)
        except:
            return sys.exc_info()[1].args[0]

    def quote(self, attr):
        return self.db.quote(attr)

    # overloadable methods serve as callbacks in the CRUD operations
    
    def before_insert_or_update(self): pass    
    def after_insert_or_update(self): pass    
    def before_insert(self): pass    
    def after_insert(self): pass    
    def before_update(self): pass    
    def after_update(self): pass    
    def before_delete(self): pass        
    def after_delete(self): pass
    def after_select(self): pass

    # import 

    @classmethod
    def from_element(Class, db, element):
        """convert an (xml) element into an instance. Rules:
        * The element attributes are primary key values
        * The element children tags are keys, texts are values
        """
        record = Class(db, **element.attrib)
        for ch in [ch for ch in element.xpath("*") if ch.text not in [None, '']]:
            tag = re.sub("^\{[^\}]*\}", "", ch.tag)
            record[tag] = ch.text
        return record




if __name__ == "__main__":
    import doctest
    doctest.testmod()



"""
A simple relational database interface that stays out of your way.
Designed to be small, fast, and transparent. Integrates with user-defined record classes.

* Record:     base class for user-defined Record classes (inherits from bl.dict.Dict)
* RecordSet:  set of Records (inherits from list)
* Database :  a database connection
                (init with a standard DB-API 2.0 connection string)

---------------------------------------------------------------------------
                                            Memory:  Footprint:
# > python                                  4656 K   4656 K (Python 3.4.3 OSX)
# >>> from bl.database import Database      5460 K    804 K (YMMV)
---------------------------------------------------------------------------
Sample session: 
>>> d = Database()      # in-memory sqlite3 database
>>> d.execute("create table table1 (name varchar primary key, email varchar not null unique)")
>>> d.execute("insert into table1 (name, email) values ('sah', 'sah@blackearthgroup.com')")
>>> records = d.select("select * from table1")
>>> records[0].name
'sah'
>>> d.connection.close()
>>>
"""

import datetime, imp, re, time, logging
from bl.dict import Dict

LOG = logging.getLogger(__name__)

class Database(Dict):
    """a database connection object."""

    def __init__(self, connection_string=None, adaptor=None, connection=None, 
                tries=3, debug=False, **args):
        Dict.__init__(self, 
            connection_string=re.sub('\s+', ' ', connection_string or ''),
            connection=connection,
            adaptor=adaptor, 
            debug=debug, tries=tries, **args)
        if self.connection is not None:
            self.adaptor = self.connection.__module__
        elif self.adaptor is None:
            import sqlite3
            self.adaptor = sqlite3
        elif type(self.adaptor) in (str, bytes):
            fm = imp.find_module(adaptor)
            self.adaptor = imp.load_module(self.adaptor, fm[0], fm[1], fm[2])
        
        # try reaching the db "tries" times, with increasing wait times, before raising an exception.
        if self.connection is None:
            for i in range(tries):
                try: 
                    if self.connection_string != None:
                         self.connection = self.adaptor.connect(self.connection_string)
                    else:
                         self.connection = self.adaptor.connect(**args)
                    break
                except: 
                    if i==list(range(tries))[-1]:       # last try failed
                        raise
                    else:                               # wait a bit
                        time.sleep(2*i)                 # doubling the time on each wait
        try:
            if self.adaptor == 'sqlite3' or 'sqlite3' in str(self.adaptor):
                self.execute("pragma foreign_keys = ON")
        except:
            pass

    def __repr__(self):
        return "Database(adaptor=%s, connection_string='%s')" % (self.adaptor, self.connection_string)

    def migrate(self, migrations=None):
        from .migration import Migration
        Migration.migrate(self, migrations=migrations or self.migrations)

    def cursor(self):
        """get a cursor for fine-grained transaction control."""
        cursor = self.connection.cursor()
        return cursor

    def execute(self, sql, vals=None, cursor=None):
        """execute SQL transaction, commit it, and return nothing. 
        If a cursor is specified, work within that transaction.
        """
        if self.debug!=False: LOG.debug("%s %r" % (sql, vals))
        try:
            c = cursor or self.connection.cursor()
            if vals in [None, (), [], {}]:
                c.execute(sql)
            else:
                c.execute(sql, vals)
            if cursor is None:
                self.commit()
        except:
            if cursor is not None:
                cursor.connection.rollback()
            raise

    def commit(self):
        """commit the changes on the current connection."""
        self.connection.commit()

    def rollback(self):
        """rollback the changes on the current connection, aborting the transaction."""
        self.connection.rollback()

    def select(self, sql, vals=None, Record=None, RecordSet=None, cursor=None):
        """select from db and return the full result set.
        Required Arguments:
            sql: the SQL query as a string
        Optional/Named Arguments
            vals: any bound variables
            Record: the class (itself) that the resulting records should be
        """
        if Record is None:
            from .record import Record
        if RecordSet is None:
            from .recordset import RecordSet

        records = RecordSet()                   # Populate a RecordSet (list) with the all resulting
        for record in self.selectgen(sql, vals=vals, Record=Record, cursor=cursor):
            records.append(record)
        return records

    def selectgen(self, sql, vals=None, Record=None, cursor=None):
        """select from db and yield a generator"""
        c = cursor or self.cursor()
        self.execute(sql, vals=vals, cursor=c)

        if Record is None:
            from .record import Record

        # get a list of attribute names from the cursor.description
        attr_list = list()
        for r in c.description:
            attr_list.append(r[0])

        result = c.fetchone()
        while result is not None:
            record = Record(self)               # whatever the record class is, include another instance
            for i in range(len(attr_list)):     # make each attribute dict-able by name
                record[attr_list[i]] = result[i]
            yield record
            result = c.fetchone()

        if cursor is None:
            c.close()   # closing the cursor without committing rolls back the transaction.

    def select_one(self, sql, vals=None, Record=None, cursor=None):
        """select one record from db
        Required Arguments:
            sql: the SQL query as a string
        Optional/Named Arguments:
            vals: any bound variables
            Record: the class (itself) that the resulting records should be
        """
        c = cursor or self.cursor()
        self.execute(sql, vals, cursor=c)

        if Record is None:
            from .record import Record

        # get a list of attribute names from the cursor.description
        attr_list = list()
        for r in c.description:
            attr_list.append(r[0])
        result = c.fetchone()
        if result is None:
            record = None
        else:
            record = Record(self)
            for i in range(len(attr_list)):
                record[attr_list[i]] = result[i]
        if cursor is None:
            c.close()
        return record

    def quote(self, attr):
        """returns the given attribute in a form that is insertable in the insert() and update() methods."""
        t = type(attr)
        if t == type(None): return 'NULL'
        elif t == datetime.datetime:    # datetime -- put it in quotes
            return "'%s'" % str(attr)
        elif t == str:
            return self._quote_str(attr)
        elif t in [dict, Dict]:
            return "$$%s$$" % attr
        else:                           # boolean or number -- no quoting needed
            return str(attr).lower()

    def _quote_str(self, attr):
        """quote the attr string in a manner fitting the Database server, if known."""
        sn = self.servername().lower()
        if 'sqlserver' in sn or 'sqlite' in sn:
            # quote for sqlserver and sqlite: double '' to escape
            attr = "'" + re.sub("'", "''", attr) + "'"
        elif 'postgres' in sn:
            if type(attr) == str:
                attr = "$$%s$$" % attr
        else:
            if type(attr) == str:
                attr = "'%s'" % attr
            else:
                attr = "'%s'" % str(attr, 'UTF-8')
        return attr

    def servername(self):
        """return a string that describes the database server being used"""
        if type(self.adaptor) in [str, bytes]:
            return self.adaptor
        elif 'psycopg' in str(self.adaptor):
            return 'postgresql'
        elif 'sqlite' in str(self.adaptor):
            return 'sqlite'
        elif self.dbconfig is not None and self.dbconfig.server is not None:
            return self.dbconfig.server
        elif 'adodbapi' in str(self.connection): 
            return 'sqlserver'
        elif 'port=5432' in str(self.connection): 
            return 'postgresql'
        elif 'port=3306' in str(self.connection): 
            return 'mysql'
        else: 
            return ''
        
    def table_names(self):
        sn = self.servername().lower()
        if 'sqlite' in sn:
            names = [r.name for r in self.select("select name from sqlite_master where type='table'")]
        elif 'mysql' in sn:
            names = [r.values()[0] for r in self.select("show tables")]
        else:
            names = [r.table_name for r in self.select("select table_name from information_schema.tables")]
        return names

    def table_exists(self, table_name):
        sn = self.servername().lower()
        if 'sqlite' in sn:
            return self.select_one(
                "select * from sqlite_master where name=? and type='table' limit 1", (table_name,))
        elif 'mysql' in sn:
            return self.select_one(
                "show tables like %s", (table_name,))
        else:   
            # postgresql and sqlserver both use the sql standard here.
            return self.select_one(
                "select * from information_schema.tables where table_name=%s limit 1", (table_name,))

def doctests():
    """
    >>> d = Database()
    >>> d.execute("create table table1 (name varchar primary key, email varchar not null unique);")
    >>> d.execute("insert into table1 (name, email) values ('sah', 'sah@blackearthgroup.com')")
    >>> d.execute("insert into table1 (name, email) values ('sah', 'sah.harrison@gmail.com')")
    Traceback (most recent call last):
      ...
    sqlite3.IntegrityError: UNIQUE constraint failed: table1.name
    """

if __name__ == "__main__":
    import doctest
    doctest.testmod()




import os, traceback, logging
from glob import glob
from bl.dict import Dict

from .model import Model

LOG = logging.getLogger(__name__)

class Migrate(Dict):
    def __init__(self, db, **Database):
        super().__init__(db=db, **Database)
    def __call__(self):
        Migration.migrate(self.db, migrations=self.migrations)

class Migration(Model):
    relation = 'migrations'
    pk = ['id']

    @classmethod
    def create_id(M, filename):
        return os.path.basename(os.path.splitext(filename)[0])

    @classmethod
    def migrate(M, db, migrations=None):
        """update the database with unapplied migrations"""
        migrations = migrations or db.migrations
        try:
            # will throw an error if this is the first migration -- migrations table doesn't yet exist.
            # (and this approach is a bit easier than querying for the existence of the table...)
            migrations_ids = [r.id for r in M(db).select()]
        except:
            migrations_ids = []
        fns = [fn for fn 
                in glob(os.path.join(migrations, "*.*")) 
                if M.create_id(fn) not in migrations_ids]
        fns.sort()
        LOG.info("Migrate Database: %d migrations" % (len(fns),))
        for fn in fns:
            id = M.create_id(fn)
            ext = os.path.splitext(fn)[1]
            if id in migrations_ids: 
                continue
            else:
                f = open(fn, 'r'); script = f.read(); f.close()
                description = script.split("\n")[0].strip('-#/*; ') # first line is the description
                LOG.info('', id+ext, ':', description)
                cursor = db.cursor()
                try:
                    if ext=='.sql':                                     # script is a SQL script, db.execute it
                        db.execute(script, cursor=cursor)
                    else:                                               # script is system script, subprocess it
                        subprocess.check_output(script, {'db': db})
                    migration = M(db, id=id, description=description)
                    migration.insert(cursor=cursor)
                    cursor.connection.commit()
                except:
                    cursor.connection.rollback()
                    raise
                cursor.close()
